// Source code Copyright (C) 2006 by Clifford A. Shaffer.

// k-d tree class
template <typename Elem> class KD {
private:
  BinNode<Elem>* root;
  int D;                 // Dimension of the key
  int nodecount;         // Number of nodes in the KD tree

  // Private "helper" functions

  // Return true iff the two points are within radius distance of each other
  bool InCircle(int* c1, int* c2, int radius) const {
    int total = 0;
    for (int i=0; i<D; i++) // Compute sum of squares
      total += (c1[i]-c2[i]) * (c1[i]-c2[i]);
    return (total <= (radius*radius));
  }

  // True iff two coords are equal
  bool EqualCoord(const int* a, const int* b) const {
    for (int i=0; i<D; i++)
      if (a[i] != b[i]) return false;
    return true;
  }

  void clearhelp(BinNode<Elem>*);
  BinNode<Elem>* inserthelp(BinNode<Elem>*, const Elem&, int);
  bool findhelp(BinNode<Elem>*, int*, Elem&, int) const;
  BinNode<Elem>* findmin(BinNode<Elem>*, int, int) const;
  BinNode<Elem>* removehelp(BinNode<Elem>*, const int*,
                            BinNode<Elem>*&, int);
  int regionhelp(BinNode<Elem>*, int*, int, int) const;
  void printhelp(BinNode<Elem>*, int) const;

public:
  // Constructor
  KD(int Din) { root = NULL; D = Din; nodecount = 0;}

  bool insert(const Elem& it) {
    root = inserthelp(root, it, 0);
    nodecount++;
    return true;
  }

  void clear()
    { clearhelp(root); root = NULL; nodecount = 0; }

  bool remove(int* coord, Elem& it) {
    BinNode<Elem>* t = NULL;
    root = removehelp(root, coord, t, 0);
    if (t == NULL) return false; // Nothing found
    it = t->val();
    nodecount--;
    delete t;
    return true;
  }

  bool find(Elem& e, int* coord) const
    { return findhelp(root, coord, e, 0); }

  void regionsearch(int* coord, int radius) const {
    int searchcount;
    searchcount = regionhelp(root, coord, radius, 0);
    cout << searchcount << " nodes visited in search\n";
  }

  int size() { return nodecount; }

  void print() const {
    if (root == NULL) cout << "The BST is empty.\n";
    else {
      printhelp(root, 0);
      cout << endl;
    }
  }
};
  
// Clean up the KD tree by releasing space back free store
template <typename Elem> void KD<Elem>::clearhelp(BinNode<Elem>* root) {
  if (root == NULL) return;
  clearhelp(root->left());
  clearhelp(root->right());
  delete root;
}

// Insert a node into the KD tree, returning the updated tree
template <typename Elem> BinNode<Elem>* KD<Elem>::
inserthelp(BinNode<Elem>* root, const Elem& it, int discrim) {
  if (root == NULL)  // Empty tree: create node
    return (new BinNodePtr<Elem>(it, NULL, NULL));
  if (((root->val())->coord())[discrim] > (it->coord())[discrim])
    root->setLeft(inserthelp(root->left(), it, (discrim+1)%D));
  else root->setRight(inserthelp(root->right(), it, (discrim+1)%D));
  return root;       // Return tree with node inserted
}

// Find the record with the given coordinates
template <typename Elem> bool KD<Elem>::
findhelp(BinNode<Elem>* root, int* coord,
         Elem& e, int discrim) const {
  // Member coord of a node is an integer array storing
  // the node's coordinates.
  if (root == NULL) return false;     // Empty tree
  int* currcoord;
  currcoord = (root->val())->coord();
  if (EqualCoord(currcoord, coord)) { // Found it
    e = root->val();
    return true;
  }
  if (currcoord[discrim] < coord[discrim])
    return findhelp(root->left(),coord,e,(discrim+1)%D);
  else
    return findhelp(root->right(),coord,e,(discrim+1)%D);
}

// Find all points within distance "rad" of location "coord"
template <typename Elem> int KD<Elem>::
regionhelp(BinNode<Elem>* root, int* coord,
           int rad, int discrim) const {
  int mycount = 0;

  if (root == NULL) return 0;     // Empty tree
  // Check if record at root is in circle
  if (InCircle((root->val())->coord(), coord, rad))
    cout << root->val() << endl;  // Do whatever is appropriate
  int* currcoord = (root->val())->coord();
  if (currcoord[discrim] > (coord[discrim] - rad))
    mycount = regionhelp(root->left(), coord, rad, (discrim+1)%D);
  if (currcoord[discrim] < (coord[discrim] + rad))
    mycount += regionhelp(root->right(), coord, rad, (discrim+1)%D);
  return (mycount + 1);
}

// Return a pointer to the node with the least value in root
// for the selected descriminator
template <typename Elem> BinNode<Elem>* KD<Elem>::
findmin(BinNode<Elem>* root,
          int discrim, int currdis) const {
  // discrim: discriminator key used for minimum search;
  // currdis: current level (mod D);
  BinNode<Elem> *temp1, *temp2;
  if (root == NULL) return NULL;
  temp1 = findmin(root->left(), discrim, (currdis+1)%D);
  if (discrim != currdis) { // If not at descrim's level,
                            // we must search both subtrees
    temp2 = findmin(root->right(),discrim,(currdis+1)%D);
    if ((temp1 == NULL) || ((temp2 != NULL) &&
         (((temp2->val())->coord())[discrim] <
	  ((temp1->val())->coord())[discrim])))
      temp1 = temp2;      // Right side has smaller key value
  } // Now, temp1 has the smallest value in children
  if ((temp1 == NULL) || (((root->val())->coord())[discrim] <
                          ((temp1->val())->coord())[discrim]))
    return root;
  else return temp1;
}


// Return in R the element (if any) which matches the coordinates K.
// Return the updated subtree with R removed from the tree.
template <typename Elem>  BinNode<Elem>* KD<Elem>::
removehelp(BinNode<Elem>* root, const int* K,
           BinNode<Elem>*& R, int descrim) {
  if (root == NULL) return NULL; // K is not in tree
  int* currcoord = (root->val())->coord();
  if (!EqualCoord(currcoord, K)) { // Continue searching
    if (K[descrim] < currcoord[descrim])
      root->setLeft(removehelp(root->left(), K, R, (descrim+1)%D));
    else root->setRight(removehelp(root->right(), K, R, (descrim+1)%D));
  }
  else {                            // Found K: remove this node
    R = root;
    if (root->right() == NULL) {    // No right child
      root->setRight(root->left()); //  so move left child to right
      root->setLeft(NULL);          // And make left NULL
    }
    if (root->right() == NULL)      // Both children are emtpy
      return NULL;                  // So the root will become NULL
    else {                          // Replace root with min on right
      BinNode<Elem>* temp;
      // First, find the min on right side
      temp = findmin(root->right(), descrim, (descrim+1)%D);
      // Now, actually get that min node out of the right side
      root->setRight(removehelp(root->right(),
		       (temp->val())->coord(), temp, (descrim+1)%D));
     // At this point, we hold the minimum node, and the children are correct
     // So just replace root with the minimum and return the old root
      Elem te = root->val();
      root->setVal(temp->val());
      temp->setVal(te);
      R = temp;
    }
  }
  return root;
}


// Print out a KD tree in preorder
template <typename Elem> void KD<Elem>::
printhelp(BinNode<Elem>* root, int level) const {

  if (root == NULL) { // Empty tree
    cout << " NULL";
    return;
  }
  // Print node value
  cout << endl;
  for (int i=0; i<level; i++)         // Indent to level
    cout << "  ";
  cout << root->val();

  // Print children
  printhelp(root->left(), level+1);   // Do left subtree
  printhelp(root->right(), level+1);  // Do right subtree
}
